@@ -1,24 +1,14 @@
-import json
-import sys
-import os
-import io
 import copy
-from collections import OrderedDict
+import json
+import os
 import warnings
-
+from dataclasses import dataclass
 from typing import Optional
 
 import numpy as np
+
 import tiledb
-from tiledb import TileDBError
-
-# from tiledb.tests.common import xprint
-
-if sys.version_info >= (3, 3):
-    unicode_type = str
-else:
-    unicode_type = unicode
-unicode_dtype = np.dtype(unicode_type)
+from tiledb import TileDBError, libtiledb
 
 
 def check_dataframe_deps():
@@ -63,6 +53,7 @@ TILEDB_KWARG_DEFAULTS = {
     "row_start_idx": None,
     "fillna": None,
     "column_types": None,
+    "varlen_types": None,
     "capacity": None,
     "date_spec": None,
     "cell_order": "row-major",
@@ -82,173 +73,142 @@ def parse_tiledb_kwargs(kwargs):
     return parsed_args
 
 
-def is_nullable_extension_type(t):
-    import pandas as pd
-
-    return isinstance(
-        t,
-        (
-            pd.Int64Dtype,
-            pd.Int32Dtype,
-            pd.Int16Dtype,
-            pd.Int8Dtype,
-            pd.UInt64Dtype,
-            pd.UInt32Dtype,
-            pd.UInt16Dtype,
-            pd.UInt8Dtype,
-        ),
-    )
-
-
+@dataclass(frozen=True)
 class ColumnInfo:
-    def __init__(self, dtype, repr: Optional[str] = None, nullable: bool = False):
-        self.dtype = dtype
-        self.repr = repr
-        self.nullable = nullable
 
+    dtype: np.dtype
+    repr: Optional[str] = None
+    nullable: bool = False
+    var: bool = False
 
-def dtype_from_column(col):
-    import pandas as pd
+    @classmethod
+    def from_values(cls, array_like, varlen_types=()):
+        from pandas.api import types as pd_types
 
-    col_dtype = col.dtype
-    if col_dtype in (
-        np.int32,
-        np.int64,
-        np.uint32,
-        np.uint64,
-        np.float,
-        np.double,
-        np.uint8,
-    ):
-        return ColumnInfo(col_dtype)
-
-    if is_nullable_extension_type(col_dtype):
-        return ColumnInfo(col_dtype.numpy_dtype, repr=str(col_dtype), nullable=True)
-
-    if isinstance(col_dtype, pd.BooleanDtype):
-        return ColumnInfo(np.uint8, repr=pd.BooleanDtype(), nullable=True)
-
-    # TODO this seems kind of brittle
-    if col_dtype.base == np.dtype("M8[ns]"):
-        if col_dtype == np.dtype("datetime64[ns]"):
-            return ColumnInfo(col_dtype)
-        elif hasattr(col_dtype, "tz"):
-            raise ValueError("datetime with tz not yet supported")
-        else:
-            raise ValueError(
-                "unsupported datetime subtype ({})".format(type(col_dtype))
-            )
-
-    # Pandas 1.0 has StringDtype extension type
-    if col_dtype.name == "string":
-        return ColumnInfo(unicode_dtype)
-
-    if col_dtype == "bool":
-        return ColumnInfo(np.uint8, repr=np.dtype("bool"))
-
-    if col_dtype == np.dtype("O"):
-        # Note: this does a full scan of the column... not sure what else to do here
-        #       because Pandas allows mixed string column types (and actually has
-        #       problems w/ allowing non-string types in object columns)
-        inferred_dtype = pd.api.types.infer_dtype(col)
-
-        if inferred_dtype == "bytes":
-            return ColumnInfo(np.bytes_)
-
-        elif inferred_dtype == "string":
-            # TODO we need to make sure this is actually convertible
-            return ColumnInfo(unicode_dtype)
-
-        elif inferred_dtype == "mixed":
-            raise ValueError(
-                "Column '{}' has mixed value dtype and cannot yet be stored as a TileDB attribute".format(
-                    col.name
-                )
-            )
-
-    raise ValueError("Unhandled column type: '{}'".format(col_dtype))
-
-
-# TODO make this a staticmethod on Attr?
-def attrs_from_df(df, index_dims=None, filters=None, column_types=None, ctx=None):
-    attr_reprs = dict()
-    if ctx is None:
-        ctx = tiledb.default_ctx()
-
-    if column_types is None:
-        column_types = dict()
-
-    attrs = list()
-    for name, col in df.items():
-        if isinstance(filters, dict):
-            if name in filters:
-                attr_filters = filters[name]
+        if pd_types.is_object_dtype(array_like):
+            # Note: this does a full scan of the column... not sure what else to do here
+            #       because Pandas allows mixed string column types (and actually has
+            #       problems w/ allowing non-string types in object columns)
+            inferred_dtype = pd_types.infer_dtype(array_like)
+            if inferred_dtype == "bytes":
+                return cls.from_dtype(np.bytes_)
+            elif inferred_dtype == "string":
+                # TODO we need to make sure this is actually convertible
+                return cls.from_dtype(np.str_)
             else:
-                attr_filters = None
-        elif filters is not None:
-            attr_filters = filters
+                raise NotImplementedError(
+                    f"{inferred_dtype} inferred dtype not supported"
+                )
         else:
-            attr_filters = tiledb.FilterList([tiledb.ZstdFilter(1, ctx=ctx)])
+            if not hasattr(array_like, "dtype"):
+                array_like = np.asanyarray(array_like)
+            return cls.from_dtype(array_like.dtype, varlen_types)
 
-        # ignore any column used as a dim/index
-        if index_dims and name in index_dims:
-            continue
-        if name in column_types:
-            spec_type = column_types[name]
-            nullable = is_nullable_extension_type(spec_type)
+    @classmethod
+    def from_dtype(cls, dtype, varlen_types=()):
+        from pandas.api import types as pd_types
 
-            actual_type = spec_type
-            # Handle ExtensionDtype
-            if hasattr(spec_type, "type"):
-                actual_type = spec_type.type
-            attr_info = ColumnInfo(actual_type, repr=str(spec_type), nullable=nullable)
+        dtype = pd_types.pandas_dtype(dtype)
+        # Note: be careful if you rearrange the order of the following checks
+
+        # extension types
+        if pd_types.is_extension_array_dtype(dtype):
+            if pd_types.is_bool_dtype(dtype):
+                np_type = np.uint8
+            else:
+                # XXX Parametrized dtypes such as "foo[int32]") sometimes have a "subtype"
+                # property that holds the "int32". If it exists use this, otherwise use
+                # the standard type property
+                np_type = getattr(dtype, "subtype", dtype.type)
+
+            var = bool(varlen_types and dtype in varlen_types)
+            if var:
+                # currently TileDB-py doesn't support nullable var-length attributes
+                nullable = False
+            else:
+                # currently nullability is a (private) property of ExtensionArray
+                # see https://github.com/pandas-dev/pandas/issues/40574
+                nullable = bool(dtype.construct_array_type()._can_hold_na)
+
+            return cls(
+                np.dtype(np_type),
+                repr=dtype.name,
+                nullable=nullable,
+                var=var,
+            )
+
+        # bool type
+        if pd_types.is_bool_dtype(dtype):
+            return cls(np.dtype("uint8"), repr="bool")
+
+        # complex types
+        if pd_types.is_complex_dtype(dtype):
+            raise NotImplementedError("complex dtype not supported")
+
+        # remaining numeric types
+        if pd_types.is_numeric_dtype(dtype):
+            if dtype == np.float16 or hasattr(np, "float128") and dtype == np.float128:
+                raise NotImplementedError(
+                    "Only single and double precision float dtypes are supported"
+                )
+            return cls(dtype)
+
+        # datetime types
+        if pd_types.is_datetime64_any_dtype(dtype):
+            if dtype == "datetime64[ns]":
+                return cls(dtype)
+            else:
+                raise NotImplementedError(
+                    "Only 'datetime64[ns]' datetime dtype is supported"
+                )
+
+        # string types
+        # don't use pd_types.is_string_dtype() because it includes object types too
+        if dtype.type in (np.bytes_, np.str_):
+            return cls(dtype)
+
+        raise NotImplementedError(f"{dtype} dtype not supported")
+
+
+def _get_column_infos(df, column_types, varlen_types):
+    column_infos = {}
+    for name, column in df.items():
+        if column_types and name in column_types:
+            column_infos[name] = ColumnInfo.from_dtype(column_types[name], varlen_types)
         else:
-            attr_info = dtype_from_column(col)
+            column_infos[name] = ColumnInfo.from_values(column, varlen_types)
+    return column_infos
 
+
+def _get_attrs(names, column_infos, attr_filters, ctx=None):
+    attrs = []
+    attr_reprs = {}
+    for name in names:
+        if isinstance(attr_filters, dict):
+            filters = attr_filters.get(name)
+        elif attr_filters is not None:
+            filters = attr_filters
+        else:
+            filters = tiledb.FilterList([tiledb.ZstdFilter(1, ctx=ctx)], ctx=ctx)
+
+        column_info = column_infos[name]
         attrs.append(
             tiledb.Attr(
                 name=name,
-                dtype=attr_info.dtype,
-                filters=attr_filters,
-                nullable=attr_info.nullable,
+                filters=filters,
+                dtype=column_info.dtype,
+                nullable=column_info.nullable,
+                var=column_info.var,
+                ctx=ctx,
             )
         )
 
-        if attr_info.repr is not None:
-            attr_reprs[name] = attr_info.repr
+        if column_info.repr is not None:
+            attr_reprs[name] = column_info.repr
 
     return attrs, attr_reprs
 
 
-def dim_info_for_column(ctx, df, col, tile=None, full_domain=False, index_dtype=None):
-
-    if isinstance(col, np.ndarray):
-        col_values = col
-    else:
-        col_values = col.values
-
-    if len(col_values) < 1:
-        raise ValueError(
-            "Empty column '{}' cannot be used for dimension!".format(col_name)
-        )
-
-    if index_dtype is not None:
-        dim_info = ColumnInfo(index_dtype)
-    elif col_values.dtype is np.dtype("O"):
-        col_val0_type = type(col_values[0])
-        if col_val0_type in (bytes, unicode_type):
-            # TODO... core only supports TILEDB_ASCII right now
-            dim_info = ColumnInfo(np.bytes_)
-        else:
-            raise TypeError(
-                "Unknown column type not yet supported ('{}')".format(col_val0_type)
-            )
-    else:
-        dim_info = dtype_from_column(col_values)
-
-    return dim_info
-
-
 def dim_for_column(
     ctx, name, dim_info, col, tile=None, full_domain=False, ndim=None, dim_filters=None
 ):
@@ -273,21 +233,16 @@ def dim_for_column(
     dtype = dim_info.dtype
 
     if full_domain:
-        if not dim_info.dtype in (np.bytes_, np.unicode_):
+        if dim_info.dtype not in (np.bytes_, np.unicode_):
             # Use the full type domain, deferring to the constructor
-            (dtype_min, dtype_max) = tiledb.libtiledb.dtype_range(dim_info.dtype)
+            dtype_min, dtype_max = tiledb.libtiledb.dtype_range(dim_info.dtype)
 
             dim_max = dtype_max
             if dtype.kind == "M":
                 date_unit = np.datetime_data(dtype)[0]
                 dim_min = np.datetime64(dtype_min + 1, date_unit)
                 tile_max = np.iinfo(np.uint64).max - tile
-
-                # modular arithmetic gives misleading overflow warning.
-                with np.errstate(over="ignore"):
-                    dtype_range = np.uint64(dtype_max) - np.uint64(dtype_min)
-
-                if np.abs(dtype_range) > tile_max:
+                if np.uint64(dtype_max - dtype_min) > tile_max:
                     dim_max = np.datetime64(dtype_max - tile, date_unit)
             elif dtype is np.int64:
                 dim_min = dtype_min + 1
@@ -296,7 +251,7 @@ def dim_for_column(
 
             if dtype.kind != "M" and np.issubdtype(dtype, np.integer):
                 tile_max = np.iinfo(np.uint64).max - tile
-                if np.abs(np.uint64(dtype_max) - np.uint64(dtype_min)) > tile_max:
+                if np.uint64(dtype_max - dtype_min) > tile_max:
                     dim_max = dtype_max - tile
         else:
             dim_min, dim_max = (None, None)
@@ -305,23 +260,26 @@ def dim_for_column(
         dim_min = np.min(col_values)
         dim_max = np.max(col_values)
 
-    if not dim_info.dtype in (np.bytes_, np.unicode_):
+    if dim_info.dtype not in (np.bytes_, np.unicode_):
         if np.issubdtype(dtype, np.integer) or dtype.kind == "M":
-            dim_range = np.uint64(np.abs(np.uint64(dim_max) - np.uint64(dim_min)))
-            # we can't make a tile larger than the dimension range
-            if dim_range < tile:
-                tile = dim_range
-            if tile < 1:
-                tile = 1
+            dim_range = np.uint64(dim_max - dim_min)
+            # we can't make a tile larger than the dimension range or lower than 1
+            tile = max(1, min(tile, dim_range))
         elif np.issubdtype(dtype, np.float64):
-            dim_range = dim_max - dim_min
+            # this difference can be inf
+            with np.errstate(over="ignore"):
+                dim_range = dim_max - dim_min
             if dim_range < tile:
                 tile = np.ceil(dim_range)
 
+    # libtiledb only supports TILEDB_ASCII dimensions, so we must use
+    # nb.bytes_ which will force encoding on write
+    dim_dtype = np.bytes_ if dim_info.dtype == np.dtype("U") else dim_info.dtype
+
     dim = tiledb.Dim(
         name=name,
         domain=(dim_min, dim_max),
-        dtype=dim_info.dtype,
+        dtype=dim_dtype,
         tile=tile,
         filters=dim_filters,
     )
@@ -336,9 +294,8 @@ def get_index_metadata(dataframe):
         if index == None:
             index_md_name = "__tiledb_rows"
         # Note: this may be expensive.
-        md[index_md_name] = dtype_from_column(
-            dataframe.index.get_level_values(index)
-        ).dtype
+        values = dataframe.index.get_level_values(index)
+        md[index_md_name] = ColumnInfo.from_values(values).dtype
 
     return md
 
@@ -348,10 +305,6 @@ def create_dims(
 ):
     import pandas as pd
 
-    index = dataframe.index
-    index_dict = OrderedDict()
-    index_dtype = None
-
     per_dim_tile = False
     if tile is not None:
         if isinstance(tile, dict):
@@ -367,47 +320,29 @@ def create_dims(
                 "Got '{}'".format(tile)
             )
 
+    index = dataframe.index
+    index_name_values = []
+    dim_types = []
+
     if isinstance(index, pd.MultiIndex):
         for name in index.names:
-            index_dict[name] = dataframe.index.get_level_values(name)
+            values = index.get_level_values(name)
+            index_name_values.append((name, values))
+            dim_types.append(ColumnInfo.from_values(values))
 
     elif isinstance(index, (pd.Index, pd.RangeIndex, pd.Int64Index)):
-        if hasattr(index, "name") and index.name is not None:
-            name = index.name
-        else:
-            index_dtype = np.dtype("uint64")
+        values = index.values
+        name = getattr(index, "name", None)
+        if name is None:
             name = "__tiledb_rows"
-
-        index_dict[name] = index.values
-
-    else:
-        raise ValueError("Unhandled index type {}".format(type(index)))
-
-    # create list of dim types
-    # we need to know all the types in order to validate before creating Dims
-    dim_types = list()
-    for idx, (name, values) in enumerate(index_dict.items()):
-        if per_dim_tile and name in tile:
-            dim_tile = tile[name]
-        elif per_dim_tile:
-            # in this case we fall back to the default
-            dim_tile = None
+            dim_types.append(ColumnInfo.from_dtype(np.dtype("uint64")))
         else:
-            # in this case we use a scalar (type-checked earlier)
-            dim_tile = tile
+            dim_types.append(ColumnInfo.from_values(values))
+        index_name_values.append((name, values))
+    else:
+        raise ValueError(f"Unhandled index type {type(index)}")
 
-        dim_types.append(
-            dim_info_for_column(
-                ctx,
-                dataframe,
-                values,
-                tile=dim_tile,
-                full_domain=full_domain,
-                index_dtype=index_dtype,
-            )
-        )
-
-    if any([d.dtype in (np.bytes_, np.unicode_) for d in dim_types]):
+    if any(d.dtype in (np.bytes_, np.unicode_) for d in dim_types):
         if sparse is False:
             raise TileDBError("Cannot create dense array with string-typed dimensions")
         elif sparse is None:
@@ -429,7 +364,10 @@ def create_dims(
     ndim = len(dim_types)
 
     dims = list()
-    for idx, (name, values) in enumerate(index_dict.items()):
+    for idx, (name, values) in enumerate(index_name_values):
+        if len(values) < 1:
+            raise ValueError(f"Empty column '{name}' cannot be used as dimension")
+
         # get the FilterList, if any
         if isinstance(filters, dict):
             if name in filters:
@@ -486,9 +424,9 @@ def create_dims(
             dims.append(
                 dim_for_column(
                     ctx,
+                    name,
                     dataframe,
                     col.values,
-                    name,
                     tile=dim_tile,
                     dim_filters=dim_filters,
                 )
@@ -511,30 +449,25 @@ def write_array_metadata(array, attr_metadata=None, index_metadata=None):
         array.meta["__pandas_index_dims"] = json.dumps(index_md_dict)
 
 
-def dataframe_to_np_arrays(dataframe, fillna=None):
-    import pandas as pd
+def _df_to_np_arrays(df, column_infos, fillna):
+    ret = {}
+    nullmaps = {}
+    for name, column in df.items():
+        column_info = column_infos[name]
 
-    if not hasattr(pd, "StringDtype"):
-        raise Exception(
-            "Unexpectedly found pandas version < 1.0; please install >= 1.0 for dataframe functionality."
-        )
+        if fillna is not None and name in fillna:
+            column = column.fillna(fillna[name])
 
-    ret = dict()
-    nullmaps = dict()
+        to_numpy_kwargs = {}
+        if not column_info.var:
+            to_numpy_kwargs.update(dtype=column_info.dtype)
 
-    for k, v in dataframe.to_dict(orient="series").items():
-        if pd.api.types.is_extension_array_dtype(v):
-            if (fillna is not None and k in fillna) and not is_nullable_extension_type(
-                v.dtype
-            ):
-                # raise ValueError("Missing 'fillna' value for column '{}' with pandas extension dtype".format(k))
-                ret[k] = v.to_numpy(na_value=fillna[k])
-            else:
-                # use default 0/empty for the dtype
-                ret[k] = v.to_numpy(dtype=v.dtype.numpy_dtype, na_value=v.dtype.type())
-                nullmaps[k] = (~v.isna()).to_numpy(dtype="uint8")
-        else:
-            ret[k] = v.to_numpy()
+        if column_info.nullable:
+            # use default 0/empty for the dtype
+            to_numpy_kwargs.update(na_value=column_info.dtype.type())
+            nullmaps[name] = (~column.isna()).to_numpy(dtype=np.uint8)
+
+        ret[name] = column.to_numpy(**to_numpy_kwargs)
 
     return ret, nullmaps
 
@@ -592,6 +525,7 @@ def from_pandas(uri, dataframe, **kwargs):
     fillna = tiledb_args.get("fillna", None)
     date_spec = tiledb_args.get("date_spec", None)
     column_types = tiledb_args.get("column_types", None)
+    varlen_types = tiledb_args.get("varlen_types", None)
 
     if mode != "append" and tiledb.array_exists(uri):
         raise TileDBError("Array URI '{}' already exists!".format(uri))
@@ -619,6 +553,16 @@ def from_pandas(uri, dataframe, **kwargs):
     if ctx is None:
         ctx = tiledb.default_ctx()
 
+    if date_spec:
+        dataframe = dataframe.assign(
+            **{
+                name: pd.to_datetime(dataframe[name], format=format)
+                for name, format in date_spec.items()
+            }
+        )
+
+    column_infos = _get_column_infos(dataframe, column_types, varlen_types)
+
     if create_array:
         if nrows:
             if full_domain is None:
@@ -638,11 +582,13 @@ def from_pandas(uri, dataframe, **kwargs):
 
         domain = tiledb.Domain(*dims, ctx=ctx)
 
-        attrs, attr_metadata = attrs_from_df(
-            dataframe,
-            index_dims=index_dims,
-            filters=attr_filters,
-            column_types=column_types,
+        attr_names = dataframe.columns
+        # ignore any column used as a dim/index
+        if index_dims:
+            attr_names = (name for name in attr_names if name not in index_dims)
+
+        attrs, attr_metadata = _get_attrs(
+            attr_names, column_infos, attr_filters, ctx=ctx
         )
 
         # don't set allows_duplicates=True for dense
@@ -664,29 +610,13 @@ def from_pandas(uri, dataframe, **kwargs):
 
         tiledb_args["mode"] = "append"
 
-    # apply fill replacements for NA values if specified
-    if fillna is not None:
-        dataframe.fillna(fillna, inplace=True)
-
-    # apply custom datetime parsing to given {'column_name': format_spec} pairs
-    # format_spec should be provied using Python format codes:
-    #     https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior
-    if date_spec is not None:
-        if type(date_spec) is not dict:
-            raise TypeError(
-                "Expected 'date_spec' to be a dict, got {}".format(type(date_spec))
-            )
-        for name, spec in date_spec.items():
-            dataframe[name] = pd.to_datetime(dataframe[name], format=spec)
-
-    # write the metadata so we can reconstruct dataframe
-    if create_array:
+        # write the metadata so we can reconstruct dataframe
         index_metadata = get_index_metadata(dataframe)
         with tiledb.open(uri, "w", ctx=ctx) as A:
             write_array_metadata(A, attr_metadata, index_metadata)
 
     if write:
-        write_dict, nullmaps = dataframe_to_np_arrays(dataframe, fillna=fillna)
+        write_dict, nullmaps = _df_to_np_arrays(dataframe, column_infos, fillna)
 
         if tiledb_args.get("debug", True):
             print("`tiledb.read_pandas` writing '{}' rows".format(len(dataframe)))
@@ -713,7 +643,7 @@ def from_pandas(uri, dataframe, **kwargs):
                         coords.append(dataframe.index.get_level_values(k))
 
                 # TODO ensure correct col/dim ordering
-                A._setitem_impl(tuple(coords), write_dict, nullmaps)
+                libtiledb._setitem_impl_sparse(A, tuple(coords), write_dict, nullmaps)
 
             else:
                 if row_start_idx is None:
@@ -724,50 +654,10 @@ def from_pandas(uri, dataframe, **kwargs):
             A.close()
 
 
-def _tiledb_result_as_dataframe(readable_array, result_dict):
-    import pandas as pd
-
-    # TODO missing key in the rep map should only be a warning, return best-effort?
-    # TODO this should be generalized for round-tripping overloadable types
-    #      for any array (e.g. np.uint8 <> bool)
-    repr_meta = None
-    index_dims = None
-    if "__pandas_attribute_repr" in readable_array.meta:
-        # backwards compatibility
-        repr_meta = json.loads(readable_array.meta["__pandas_attribute_repr"])
-    if "__pandas_index_dims" in readable_array.meta:
-        index_dims = json.loads(readable_array.meta["__pandas_index_dims"])
-
-    indexes = list()
-    rename_cols = dict()
-
-    for col_name, col_val in result_dict.items():
-        if repr_meta and col_name in repr_meta:
-            new_col = pd.Series(col_val, dtype=repr_meta[col_name])
-            result_dict[col_name] = new_col
-        elif index_dims and col_name in index_dims:
-            new_col = pd.Series(col_val, dtype=index_dims[col_name])
-            result_dict[col_name] = new_col
-            if col_name == "__tiledb_rows":
-                rename_cols["__tiledb_rows"] = None
-                indexes.append(None)
-            else:
-                indexes.append(col_name)
-
-    for col_key, col_name in rename_cols.items():
-        result_dict[col_name] = result_dict.pop(col_key)
-
-    df = pd.DataFrame.from_dict(result_dict)
-    if len(indexes) > 0:
-        df.set_index(indexes, inplace=True)
-
-    return df
-
-
-def open_dataframe(uri, ctx=None):
+def open_dataframe(uri, *, attrs=None, use_arrow=None, idx=slice(None), ctx=None):
     """Open TileDB array at given URI as a Pandas dataframe
 
-    If the array was saved using tiledb.from_dataframe, then columns
+    If the array was saved using tiledb.from_pandas, then columns
     will be interpreted as non-primitive pandas or numpy types when
     available.
 
@@ -783,15 +673,14 @@ def open_dataframe(uri, ctx=None):
     """
     check_dataframe_deps()
 
-    if ctx is None:
-        ctx = tiledb.default_ctx()
     # TODO support `distributed=True` option?
     with tiledb.open(uri, ctx=ctx) as A:
-        nonempty = A.nonempty_domain()
-        data = A.multi_index.__getitem__(tuple(slice(s1, s2) for s1, s2 in nonempty))
-        new_df = _tiledb_result_as_dataframe(A, data)
+        df = A.query(attrs=attrs, use_arrow=use_arrow, coords=True).df[idx]
 
-    return new_df
+    if attrs and list(df.columns) != list(attrs):
+        df = df[attrs]
+
+    return df
 
 
 def _iterate_csvs_pandas(csv_list, pandas_args):
