@@ -1,21 +1,124 @@
-import tiledb
-from tiledb.tests.common import DiskTestCase
-
 import itertools
+
 import numpy as np
+import pytest
+
+import tiledb
+from tiledb import _fragment
+from tiledb.tests.common import DiskTestCase
 
 
 class FragmentInfoTest(DiskTestCase):
     def setUp(self):
-        super(FragmentInfoTest, self).setUp()
+        super().setUp()
         if not tiledb.libtiledb.version() >= (2, 2):
-            self.skipTest("Only run FragmentInfo test with TileDB>=2.2")
+            pytest.skip("Only run FragmentInfo test with TileDB>=2.2")
 
     def test_uri_dne(self):
-        fragment_info = tiledb.fragment_info("does_not_exist")
+        fragment_info = _fragment.info("does_not_exist")
         with self.assertRaises(tiledb.TileDBError):
             fragment_info.load()
 
+    def test_array_fragments(self):
+        fragments = 3
+
+        A = np.zeros(fragments)
+
+        uri = self.path("test_dense_fragments")
+        ctx = tiledb.default_ctx()
+        dom = tiledb.Domain(
+            tiledb.Dim(ctx=ctx, domain=(0, 2), tile=fragments, dtype=np.int64), ctx=ctx
+        )
+        att = tiledb.Attr(ctx=ctx, dtype=A.dtype)
+        schema = tiledb.ArraySchema(ctx=ctx, domain=dom, attrs=(att,))
+
+        tiledb.DenseArray.create(uri, schema)
+
+        for fragment_idx in range(fragments):
+            timestamp = fragment_idx + 1
+            with tiledb.DenseArray(uri, mode="w", timestamp=timestamp, ctx=ctx) as T:
+                T[fragment_idx : fragment_idx + 1] = fragment_idx
+
+        fragments_info = tiledb.array_fragments(uri, ctx)
+
+        self.assertEqual(len(fragments_info), 3)
+        self.assertEqual(fragments_info.unconsolidated_metadata_num, 3)
+
+        self.assertEqual(fragments_info.cell_num, (3, 3, 3))
+        self.assertEqual(fragments_info.dense, (True, True, True))
+        self.assertEqual(
+            fragments_info.has_consolidated_metadata, (False, False, False)
+        )
+        self.assertEqual(
+            fragments_info.non_empty_domain, (((0, 0),), ((1, 1),), ((2, 2),))
+        )
+        self.assertEqual(fragments_info.sparse, (False, False, False))
+        self.assertEqual(fragments_info.timestamp_range, ((1, 1), (2, 2), (3, 3)))
+        self.assertEqual(fragments_info.to_vacuum_num, 0)
+        self.assertEqual(fragments_info.to_vacuum_uri, [])
+
+        if tiledb.libtiledb.version() < (2, 2, 3):
+            self.assertEqual(fragments_info.version, (7, 7, 7))
+        elif tiledb.libtiledb.version() < (2, 3, 0):
+            self.assertEqual(fragments_info.version, (8, 8, 8))
+        else:
+            self.assertEqual(fragments_info.version, (9, 9, 9))
+
+        for idx, frag in enumerate(fragments_info):
+            self.assertEqual(frag.cell_num, 3)
+            self.assertEqual(frag.dense, True)
+            self.assertEqual(frag.has_consolidated_metadata, False)
+            self.assertEqual(frag.non_empty_domain, ((idx, idx),))
+            self.assertEqual(frag.sparse, False)
+            self.assertEqual(frag.timestamp_range, (idx + 1, idx + 1))
+
+            if tiledb.libtiledb.version() < (2, 2, 3):
+                self.assertEqual(frag.version, 7)
+            elif tiledb.libtiledb.version() < (2, 3, 0):
+                self.assertEqual(frag.version, 8)
+            else:
+                self.assertEqual(frag.version, 9)
+
+    def test_array_fragments_var(self):
+        fragments = 3
+
+        uri = self.path("test_array_fragments_var")
+        dom = tiledb.Domain(
+            tiledb.Dim(name="dim", domain=(None, None), tile=None, dtype=np.bytes_)
+        )
+        schema = tiledb.ArraySchema(
+            domain=dom,
+            sparse=True,
+            attrs=[tiledb.Attr(name="1s", dtype=np.int32, var=True)],
+        )
+
+        tiledb.SparseArray.create(uri, schema)
+
+        for fragment_idx in range(fragments):
+            timestamp = fragment_idx + 1
+
+            data = np.array(
+                [
+                    np.array([timestamp] * 1, dtype=np.int32),
+                    np.array([timestamp] * 2, dtype=np.int32),
+                    np.array([timestamp] * 3, dtype=np.int32),
+                ],
+                dtype="O",
+            )
+
+            with tiledb.SparseArray(uri, mode="w", timestamp=timestamp) as T:
+                T[["zero", "one", "two"]] = data
+
+        fragments_info = tiledb.array_fragments(uri)
+
+        self.assertEqual(
+            fragments_info.non_empty_domain,
+            ((("one", "zero"),), (("one", "zero"),), (("one", "zero"),)),
+        )
+
+        for frag in fragments_info:
+            self.assertEqual(frag.non_empty_domain, (("one", "zero"),))
+
     def test_dense_fragments(self):
         fragments = 3
 
@@ -31,7 +134,7 @@ class FragmentInfoTest(DiskTestCase):
 
         tiledb.DenseArray.create(uri, schema)
 
-        fragment_info = tiledb.fragment_info(uri)
+        fragment_info = _fragment.info(uri)
 
         for fragment_idx in range(fragments):
             timestamp = fragment_idx + 1
@@ -78,8 +181,10 @@ class FragmentInfoTest(DiskTestCase):
         self.assertEqual(fragment_info.sparse(), (False, False, False))
         if tiledb.libtiledb.version() < (2, 2, 3):
             self.assertEqual(fragment_info.version(), (7, 7, 7))
-        else:
+        elif tiledb.libtiledb.version() < (2, 3, 0):
             self.assertEqual(fragment_info.version(), (8, 8, 8))
+        else:
+            self.assertEqual(fragment_info.version(), (9, 9, 9))
 
     def test_sparse_fragments(self):
         fragments = 3
@@ -96,7 +201,7 @@ class FragmentInfoTest(DiskTestCase):
 
         tiledb.SparseArray.create(uri, schema)
 
-        fragment_info = tiledb.fragment_info(uri)
+        fragment_info = _fragment.info(uri)
 
         for fragment_idx in range(fragments):
             timestamp = fragment_idx + 1
@@ -143,8 +248,10 @@ class FragmentInfoTest(DiskTestCase):
         self.assertEqual(fragment_info.sparse(), (True, True, True))
         if tiledb.libtiledb.version() < (2, 2, 3):
             self.assertEqual(fragment_info.version(), (7, 7, 7))
-        else:
+        elif tiledb.libtiledb.version() < (2, 3, 0):
             self.assertEqual(fragment_info.version(), (8, 8, 8))
+        else:
+            self.assertEqual(fragment_info.version(), (9, 9, 9))
 
     def test_non_empty_domain(self):
         uri = self.path("test_non_empty_domain")
@@ -172,7 +279,7 @@ class FragmentInfoTest(DiskTestCase):
             y = [-1.5, -1.25]
             T[x, y] = np.array(range(2))
 
-        fragment_info = tiledb.fragment_info(uri, ctx)
+        fragment_info = _fragment.info(uri, ctx)
         fragment_info.load()
 
         x_dt = schema.domain.dim(0).dtype
@@ -236,7 +343,7 @@ class FragmentInfoTest(DiskTestCase):
             )
             T[dates] = np.array(range(3))
 
-        fragment_info = tiledb.fragment_info(uri, ctx)
+        fragment_info = _fragment.info(uri, ctx)
         fragment_info.load()
 
         self.assertEqual(
@@ -297,28 +404,28 @@ class FragmentInfoTest(DiskTestCase):
             y_dims = [b"e", b"f"]
             T[x_dims, y_dims] = np.array([1, 2])
 
-        fragment_info = tiledb.fragment_info(uri, ctx)
+        fragment_info = _fragment.info(uri, ctx)
         fragment_info.load()
 
-        self.assertEqual(fragment_info.non_empty_domain_var(schema, 0, 0), ("a", "d"))
-        self.assertEqual(fragment_info.non_empty_domain_var(schema, 0, 1), ("e", "h"))
-        self.assertEqual(fragment_info.non_empty_domain_var(schema, 1, 0), ("a", "b"))
-        self.assertEqual(fragment_info.non_empty_domain_var(schema, 1, 1), ("e", "f"))
+        self.assertEqual(fragment_info.get_non_empty_domain(schema, 0, 0), ("a", "d"))
+        self.assertEqual(fragment_info.get_non_empty_domain(schema, 0, 1), ("e", "h"))
+        self.assertEqual(fragment_info.get_non_empty_domain(schema, 1, 0), ("a", "b"))
+        self.assertEqual(fragment_info.get_non_empty_domain(schema, 1, 1), ("e", "f"))
 
-        self.assertEqual(fragment_info.non_empty_domain_var(schema, 0, "x"), ("a", "d"))
-        self.assertEqual(fragment_info.non_empty_domain_var(schema, 0, "y"), ("e", "h"))
-        self.assertEqual(fragment_info.non_empty_domain_var(schema, 1, "x"), ("a", "b"))
-        self.assertEqual(fragment_info.non_empty_domain_var(schema, 1, "y"), ("e", "f"))
+        self.assertEqual(fragment_info.get_non_empty_domain(schema, 0, "x"), ("a", "d"))
+        self.assertEqual(fragment_info.get_non_empty_domain(schema, 0, "y"), ("e", "h"))
+        self.assertEqual(fragment_info.get_non_empty_domain(schema, 1, "x"), ("a", "b"))
+        self.assertEqual(fragment_info.get_non_empty_domain(schema, 1, "y"), ("e", "f"))
 
         self.assertEqual(
-            fragment_info.non_empty_domain_var(schema, 0), (("a", "d"), ("e", "h"))
+            fragment_info.get_non_empty_domain(schema, 0), (("a", "d"), ("e", "h"))
         )
         self.assertEqual(
-            fragment_info.non_empty_domain_var(schema, 1), (("a", "b"), ("e", "f"))
+            fragment_info.get_non_empty_domain(schema, 1), (("a", "b"), ("e", "f"))
         )
 
         self.assertEqual(
-            fragment_info.non_empty_domain_var(schema),
+            fragment_info.get_non_empty_domain(schema),
             ((("a", "d"), ("e", "h")), (("a", "b"), ("e", "f"))),
         )
 
@@ -331,7 +438,7 @@ class FragmentInfoTest(DiskTestCase):
 
         tiledb.SparseArray.create(uri, schema)
 
-        fragment_info = tiledb.fragment_info(uri, ctx)
+        fragment_info = _fragment.info(uri, ctx)
 
         with tiledb.SparseArray(uri, mode="w", ctx=ctx) as T:
             a = np.array([1, 2, 3, 4])
@@ -341,7 +448,7 @@ class FragmentInfoTest(DiskTestCase):
             b = np.array([1, 2])
             T[b] = b
 
-        fragment_info = tiledb.fragment_info(uri, ctx)
+        fragment_info = _fragment.info(uri, ctx)
         fragment_info.load()
 
         self.assertEqual(fragment_info.cell_num(0), len(a))
@@ -361,7 +468,7 @@ class FragmentInfoTest(DiskTestCase):
 
         tiledb.DenseArray.create(uri, schema)
 
-        fragment_info = tiledb.fragment_info(uri, ctx)
+        fragment_info = _fragment.info(uri, ctx)
 
         for fragment_idx in range(fragments):
             with tiledb.DenseArray(uri, mode="w", ctx=ctx) as T:
@@ -404,7 +511,7 @@ class FragmentInfoTest(DiskTestCase):
 
         tiledb.DenseArray.create(uri, schema)
 
-        fragment_info = tiledb.fragment_info(uri, ctx)
+        fragment_info = _fragment.info(uri, ctx)
 
         for fragment_idx in range(fragments):
             with tiledb.DenseArray(uri, mode="w", ctx=ctx) as T:
