@@ -1,27 +1,28 @@
-from __future__ import absolute_import
+import pytest
 
-try:
-    import pandas as pd
-    import pandas._testing as tm
+pd = pytest.importorskip("pandas")
+tm = pd._testing
 
-    import_failed = False
-except ImportError:
-    import_failed = True
-
-import unittest, os
-import warnings
-import string, random, copy
-import numpy as np
-from numpy.testing import assert_array_equal, assert_array_almost_equal
+import copy
+import glob
+import os
+import random
 from pathlib import Path
 
-import tiledb
-from tiledb.tests.common import *
+import numpy as np
+from numpy.testing import assert_array_equal
 
-if sys.version_info > (3, 0):
-    str_type = str
-else:
-    str_type = unicode
+import tiledb
+from tiledb.dataframe_ import ColumnInfo
+from tiledb.tests.common import (
+    DiskTestCase,
+    checked_path,
+    dtype_max,
+    rand_ascii,
+    rand_ascii_bytes,
+    rand_datetime64_array,
+    rand_utf8,
+)
 
 
 def make_dataframe_basic1(col_size=10):
@@ -64,9 +65,6 @@ def make_dataframe_basic2():
     #   https://github.com/pandas-dev/pandas/blob/master/pandas/tests/io/test_feather.py
     # (available under BSD 3-clause license
     #   https://github.com/pandas-dev/pandas/blob/master/LICENSE
-
-    import pandas as pd
-
     df = pd.DataFrame(
         {
             "string": list("abc"),
@@ -105,13 +103,140 @@ def make_dataframe_basic3(col_size=10, time_range=(None, None)):
     return df
 
 
-class PandasDataFrameRoundtrip(DiskTestCase):
-    def setUp(self):
-        if import_failed:
-            self.skipTest("Pandas not available")
-        else:
-            super(PandasDataFrameRoundtrip, self).setUp()
+class TestColumnInfo:
+    def assertColumnInfo(self, info, info_dtype, info_repr=None, info_nullable=False):
+        assert isinstance(info.dtype, np.dtype)
+        assert info.dtype == info_dtype
 
+        assert info.repr is None or isinstance(info.repr, str)
+        assert info.repr == info_repr
+
+        assert isinstance(info.nullable, bool)
+        assert info.nullable == info_nullable
+
+    @pytest.mark.parametrize(
+        "type_specs, info_dtype, info_repr, info_nullable",
+        [
+            # bool types
+            ([bool, "b1"], np.dtype("uint8"), "bool", False),
+            ([pd.BooleanDtype()], np.dtype("uint8"), "boolean", True),
+            # numeric types
+            ([np.uint8, "u1"], np.dtype("uint8"), None, False),
+            ([np.uint16, "u2"], np.dtype("uint16"), None, False),
+            ([np.uint32, "u4"], np.dtype("uint32"), None, False),
+            ([np.uint64, "u8"], np.dtype("uint64"), None, False),
+            ([np.int8, "i1"], np.dtype("int8"), None, False),
+            ([np.int16, "i2"], np.dtype("int16"), None, False),
+            ([np.int32, "i4"], np.dtype("int32"), None, False),
+            ([np.int64, "i8"], np.dtype("int64"), None, False),
+            ([np.float32, "f4"], np.dtype("float32"), None, False),
+            ([np.float64, "f8", float], np.dtype("float64"), None, False),
+            # nullable int types
+            ([pd.UInt8Dtype(), "UInt8"], np.dtype("uint8"), "UInt8", True),
+            ([pd.UInt16Dtype(), "UInt16"], np.dtype("uint16"), "UInt16", True),
+            ([pd.UInt32Dtype(), "UInt32"], np.dtype("uint32"), "UInt32", True),
+            ([pd.UInt64Dtype(), "UInt64"], np.dtype("uint64"), "UInt64", True),
+            ([pd.Int8Dtype(), "Int8"], np.dtype("int8"), "Int8", True),
+            ([pd.Int16Dtype(), "Int16"], np.dtype("int16"), "Int16", True),
+            ([pd.Int32Dtype(), "Int32"], np.dtype("int32"), "Int32", True),
+            ([pd.Int64Dtype(), "Int64"], np.dtype("int64"), "Int64", True),
+            # datetime types
+            (["datetime64[ns]"], np.dtype("<M8[ns]"), None, False),
+            # string types
+            ([np.str_, str], np.dtype("<U"), None, False),
+            ([np.bytes_, bytes], np.dtype("S"), None, False),
+            ([pd.StringDtype()], np.dtype("<U"), "string", True),
+        ],
+    )
+    def test_implemented(self, type_specs, info_dtype, info_repr, info_nullable):
+        assert isinstance(info_dtype, np.dtype)
+        assert info_repr is None or isinstance(info_repr, str)
+        assert isinstance(info_nullable, bool)
+        for type_spec in type_specs:
+            self.assertColumnInfo(
+                ColumnInfo.from_dtype(type_spec), info_dtype, info_repr, info_nullable
+            )
+
+            series = pd.Series([], dtype=type_spec)
+            if series.dtype == type_spec:
+                self.assertColumnInfo(
+                    ColumnInfo.from_values(series), info_dtype, info_repr, info_nullable
+                )
+
+    def test_object_dtype(self):
+        self.assertColumnInfo(
+            ColumnInfo.from_values(pd.Series(["hello", "world"])), np.dtype("<U")
+        )
+        self.assertColumnInfo(
+            ColumnInfo.from_values(pd.Series([b"hello", b"world"])), np.dtype("S")
+        )
+        for s in ["hello", b"world"], ["hello", 1], [b"hello", 1]:
+            pytest.raises(NotImplementedError, ColumnInfo.from_values, pd.Series(s))
+
+    unsupported_type_specs = [
+        [np.float16, "f2"],
+        [np.complex64, "c8"],
+        [np.complex128, "c16"],
+        [np.datetime64, "<M8", "datetime64"],
+        [
+            "<M8[Y]",
+            "<M8[M]",
+            "<M8[W]",
+            "<M8[D]",
+            "<M8[h]",
+            "<M8[m]",
+            "<M8[s]",
+            "<M8[ms]",
+            "<M8[us]",
+            "<M8[ps]",
+            "<M8[fs]",
+            "<M8[as]",
+        ],
+    ]
+    if hasattr(np, "float128"):
+        unsupported_type_specs.append([np.float128, "f16"])
+    if hasattr(np, "complex256"):
+        unsupported_type_specs.append([np.complex256, "c32"])
+
+    @pytest.mark.parametrize("type_specs", unsupported_type_specs)
+    def test_not_implemented(self, type_specs):
+        for type_spec in type_specs:
+            pytest.raises(NotImplementedError, ColumnInfo.from_dtype, type_spec)
+            try:
+                series = pd.Series([], dtype=type_spec)
+            except (ValueError, TypeError):
+                pass
+            else:
+                if series.dtype == type_spec:
+                    pytest.raises(NotImplementedError, ColumnInfo.from_values, series)
+
+
+class TestDimType:
+    @pytest.mark.parametrize(
+        ["df", "expected", "kwargs"],
+        [
+            (
+                pd.DataFrame(
+                    {"data": np.array(["a"], dtype=np.str_)},
+                    index=pd.Index(np.array(["b"], dtype=np.str_), name="str_dim"),
+                ),
+                np.bytes_,
+                {},
+            )
+        ],
+    )
+    def test_schema_dim(self, checked_path, df, expected, kwargs):
+        assert isinstance(df, pd.DataFrame)
+
+        uri = checked_path.path()
+        tiledb.from_pandas(uri, df, **kwargs)
+
+        with tiledb.open(uri) as A:
+            assert A.schema.domain.dim(0).dtype == expected
+            assert A.schema.domain.dim(0).name == df.index.name
+
+
+class TestPandasDataFrameRoundtrip(DiskTestCase):
     def test_dataframe_basic_rt1_manual(self):
 
         uri = self.path("dataframe_basic_rt1_manual")
@@ -179,32 +304,39 @@ class PandasDataFrameRoundtrip(DiskTestCase):
         df = make_dataframe_basic1()
 
         ctx = tiledb.Ctx()
-        tiledb.from_dataframe(uri, df, sparse=False, ctx=ctx)
+        tiledb.from_pandas(uri, df, sparse=False, ctx=ctx)
 
         df_readback = tiledb.open_dataframe(uri)
         tm.assert_frame_equal(df, df_readback)
 
         uri = self.path("dataframe_basic_rt1_unlimited")
-        tiledb.from_dataframe(uri, df, full_domain=True, sparse=False, ctx=ctx)
-        df_readback = tiledb.open_dataframe(uri)
-        tm.assert_frame_equal(df, df_readback)
+        tiledb.from_pandas(uri, df, full_domain=True, sparse=False, ctx=ctx)
 
-        with tiledb.open(uri) as A:
-            df_arrow = A.query(use_arrow=True).df[:]
-        tm.assert_frame_equal(df, df_arrow)
+        for use_arrow in None, False, True:
+            df_readback = tiledb.open_dataframe(uri, use_arrow=use_arrow)
+            tm.assert_frame_equal(df, df_readback)
+
+            attrs = ["s", "q", "t"]
+            df_readback = tiledb.open_dataframe(uri, attrs=attrs, use_arrow=use_arrow)
+            tm.assert_frame_equal(df[attrs], df_readback)
+
+            df_readback = tiledb.open_dataframe(
+                uri, idx=slice(2, 4), use_arrow=use_arrow
+            )
+            tm.assert_frame_equal(df[2:5], df_readback)
 
     def test_dataframe_basic2(self):
         uri = self.path("dataframe_basic_rt2")
 
         df = make_dataframe_basic2()
 
-        tiledb.from_dataframe(uri, df, sparse=False)
+        tiledb.from_pandas(uri, df, sparse=False)
 
         df_readback = tiledb.open_dataframe(uri)
         tm.assert_frame_equal(df, df_readback)
 
         with tiledb.open(uri) as B:
-            tm.assert_frame_equal(df, B.df[:], check_index_type=False)
+            tm.assert_frame_equal(df, B.df[:])
 
     def test_dataframe_csv_rt1(self):
         def rand_dtype(dtype, size):
@@ -295,13 +427,13 @@ class PandasDataFrameRoundtrip(DiskTestCase):
 
                 # also ensure that string columns are converted to bytes
                 # b/c only TILEDB_ASCII supported for string dimension
-                if type(df[col][0]) == str_type:
+                if type(df[col][0]) is str:
                     df[col] = [x.encode("UTF-8") for x in df[col]]
 
             new_df = df.drop_duplicates(subset=col)
             new_df.set_index(col, inplace=True)
 
-            tiledb.from_dataframe(uri, new_df, sparse=True)
+            tiledb.from_pandas(uri, new_df, sparse=True)
 
             with tiledb.open(uri) as A:
                 self.assertEqual(A.domain.dim(0).name, col)
@@ -321,7 +453,7 @@ class PandasDataFrameRoundtrip(DiskTestCase):
         df_dict = df.to_dict(orient="series")
         df.set_index(["time", "double_range"], inplace=True)
 
-        tiledb.from_dataframe(uri, df, sparse=True)
+        tiledb.from_pandas(uri, df, sparse=True)
 
         with tiledb.open(uri) as A:
             ned_time = A.nonempty_domain()[0]
@@ -345,6 +477,69 @@ class PandasDataFrameRoundtrip(DiskTestCase):
             df_idx_res = A.query(use_arrow=False).df[slice(*ned_time), :]
             tm.assert_frame_equal(df_idx_res, df)
 
+    def test_dataframe_index_name(self):
+        uri = self.path("df_index_name")
+
+        df = make_dataframe_basic1(10)
+        df.index.name = "range_idx"
+
+        tiledb.from_pandas(uri, df)
+
+        df_readback = tiledb.open_dataframe(uri, use_arrow=False)
+        tm.assert_frame_equal(df, df_readback)
+
+        # TODO: fix index name when using pyarrow
+        # df_readback = tiledb.open_dataframe(uri, use_arrow=True)
+        # tm.assert_frame_equal(df, df_readback)
+
+    def test_dataframe_fillna(self):
+        data = pd.Series(np.arange(10), dtype=pd.Int64Dtype())
+        data[4] = None
+        df = pd.DataFrame({"v": data})
+        df_copy = df.copy()
+
+        uri = self.path("df_fillna")
+        tiledb.from_pandas(uri, df, fillna={"v": -1})
+
+        # check that original df has not changed
+        tm.assert_frame_equal(df, df_copy)
+
+        # update the value in the original dataframe to match what we expect on read-back
+        df["v"][4] = -1
+        df_bk = tiledb.open_dataframe(uri)
+        tm.assert_frame_equal(df_bk, df)
+
+    def test_dataframe_date_spec(self):
+        date_fmt = "%d/%m/%Y"
+        df = make_dataframe_basic3()
+        df["date"] = df["time"].dt.strftime(date_fmt)
+
+        df_copy = df.copy()
+
+        uri = self.path("df_date_spec")
+        tiledb.from_pandas(uri, df, date_spec={"date": date_fmt})
+
+        # check that original df has not changed
+        tm.assert_frame_equal(df, df_copy)
+
+        # update the column in the original dataframe to match what we expect on read-back
+        df["date"] = pd.to_datetime(df["date"], format=date_fmt)
+        df_bk = tiledb.open_dataframe(uri)
+        tm.assert_frame_equal(df_bk, df)
+
+    def test_dataframe_empty(self):
+        dfs = [
+            make_dataframe_basic1(),
+            make_dataframe_basic2(),
+            make_dataframe_basic3(),
+        ]
+        for i, df in enumerate(dfs, start=1):
+            for sparse in False, True:
+                uri = self.path(f"dataframe_empty_{i}_{sparse}")
+                tiledb.from_pandas(uri, df, sparse=sparse)
+                with tiledb.open(uri) as A:
+                    tm.assert_frame_equal(df.iloc[:0], A.df[tiledb.EmptyRange])
+
     def test_csv_dense(self):
         col_size = 10
         df_data = {
@@ -454,7 +649,7 @@ class PandasDataFrameRoundtrip(DiskTestCase):
         df.to_csv(tmp_csv, index=False)
 
         attrs_filters = tiledb.FilterList([tiledb.ZstdFilter(1)])
-        # from_dataframe default is 1, so use 7 here to check
+        # from_pandas default is 1, so use 7 here to check
         #   the arg is correctly parsed/passed
         coords_filters = tiledb.FilterList([tiledb.ZstdFilter(7)])
 
@@ -589,7 +784,7 @@ class PandasDataFrameRoundtrip(DiskTestCase):
 
             # test .df[] indexing
             df_idx_res = A.df[int(ned[0]) : int(ned[1])]
-            tm.assert_frame_equal(df_idx_res, df, check_index_type=False)
+            tm.assert_frame_equal(df_idx_res, df)
 
             # test .df[] indexing with query
             df_idx_res = A.query(attrs=["time"]).df[int(ned[0]) : int(ned[1])]
@@ -617,9 +812,8 @@ class PandasDataFrameRoundtrip(DiskTestCase):
             # update the value in the original dataframe to match what we expect on read-back
             df["v"][4] = 0
 
-            with tiledb.open(path) as A:
-                df_bk = A.df[:]
-                tm.assert_frame_equal(df_bk, df, check_index_type=False)
+            df_bk = tiledb.open_dataframe(path)
+            tm.assert_frame_equal(df_bk, df)
 
         ### Test 1
         col_size = 10
@@ -650,7 +844,6 @@ class PandasDataFrameRoundtrip(DiskTestCase):
                 tmp_array2,
                 tmp_csv2,
                 fillna={"v": 0},
-                dtype={"v": pd.Int64Dtype()},
                 column_types={"v": pd.Int64Dtype()},
                 sparse=True,
             )
@@ -750,7 +943,7 @@ class PandasDataFrameRoundtrip(DiskTestCase):
         df = make_dataframe_basic3(col_size)
         df.set_index(["time"], inplace=True)
 
-        tiledb.from_dataframe(uri, df, sparse=True)
+        tiledb.from_pandas(uri, df, sparse=True)
 
         with tiledb.open(uri) as A:
             with self.assertRaises(tiledb.TileDBError):
@@ -797,9 +990,8 @@ class PandasDataFrameRoundtrip(DiskTestCase):
             )
 
             tiledb.from_parquet(str(tdb_uri), str(pq_uri))
-
-            with tiledb.open(tdb_uri) as T:
-                tm.assert_frame_equal(df, T.df[:], check_index_type=False)
+            df_bk = tiledb.open_dataframe(tdb_uri)
+            tm.assert_frame_equal(df, df_bk)
 
         basic1 = make_dataframe_basic1()
         try_rt("basic1", basic1)
@@ -838,8 +1030,8 @@ class PandasDataFrameRoundtrip(DiskTestCase):
 
             tiledb.from_pandas(uri, df)
 
-            with tiledb.open(uri) as A:
-                tm.assert_frame_equal(df, A.df[:], check_index_type=False)
+            df_bk = tiledb.open_dataframe(uri)
+            tm.assert_frame_equal(df, df_bk)
 
     def test_nullable_bool(self):
         uri = self.path("test_nullable_bool")
@@ -855,5 +1047,36 @@ class PandasDataFrameRoundtrip(DiskTestCase):
 
         tiledb.from_pandas(uri, df)
 
+        df_bk = tiledb.open_dataframe(uri)
+        tm.assert_frame_equal(df, df_bk)
+
+    def test_var_length(self):
+        from tiledb.tests.datatypes import RaggedDtype
+
+        dtype = np.dtype("uint16")
+        data = np.empty(100, dtype="O")
+        data[:] = [
+            np.random.randint(1000, size=np.random.randint(1, 10), dtype=dtype)
+            for _ in range(len(data))
+        ]
+
+        df = pd.DataFrame({"data": data})
+
+        uri = self.path("test_var_length")
+        data_dtype = RaggedDtype(dtype)
+        tiledb.from_pandas(
+            uri, df, column_types={"data": data_dtype}, varlen_types={data_dtype}
+        )
+
         with tiledb.open(uri) as A:
-            tm.assert_frame_equal(df, A.df[:], check_index_type=False)
+            # TODO: update the test when we support Arrow lists
+            import pyarrow
+
+            with pytest.raises(pyarrow.lib.ArrowInvalid):
+                A.df[:]
+
+            df2 = A.query(use_arrow=False).df[:]
+            tm.assert_frame_equal(df, df2, check_dtype=False)
+            for array1, array2 in zip(df["data"].values, df2["data"].values):
+                self.assertEqual(array1.dtype, array2.dtype)
+                np.testing.assert_array_equal(array1, array2)
